// -------------------------------------------------------------------
// MAdLib - Copyright (C) 2008-2009 Universite catholique de Louvain
//
// See the Copyright.txt and License.txt files for license information. 
// You should have received a copy of these files along with MAdLib. 
// If not, see <http://www.madlib.be/license/>
//
// Please report all bugs and problems to <contrib@madlib.be>
//
// Authors: Gaetan Compere, Jean-Francois Remacle
// -------------------------------------------------------------------

#include "AnisoMeshSize.h"
#include "MathUtils.h"
#include "MAdMessage.h"
#include "MAdMatrix.h"
#include "MAdVector3.h"
#include "MAdDefines.h"

#include <iostream>
using std::string;

namespace MAd {

  // -------------------------------------------------------------------
  AnisoMeshSize::AnisoMeshSize(double dirs[3][3], double _h[3]):
    MeshSizeBase()
  {
    if( _h[0] <= 0. || _h[1] <= 0. || _h[2] <= 0. ) {
      MAdMsgSgl::instance().error(__LINE__,__FILE__,
                                  "Negative size(s): %f %f %f",
                                  _h[0], _h[1], _h[2]);
    }

    vec3 t1 = vec3();
    vec3 t2 = vec3();
    vec3 t3 = vec3();
    double l1,l2,l3;

    // sort the sizes from lowest to highest and fill first direction (GCRemark: not useful)
    int hMin = indexOfMin(_h[0], _h[1], _h[2]);
    l1 = 1. / (_h[hMin] * _h[hMin]);
    double normInv = 1. / sqrt ( dirs[hMin][0] * dirs[hMin][0] + 
                                 dirs[hMin][1] * dirs[hMin][1] + 
                                 dirs[hMin][2] * dirs[hMin][2]   );
    for (int i=0; i<3; i++) t1(i) = dirs[hMin][i] * normInv;

    hMin++;
    if( _h[(hMin+1)%3] > _h[(hMin)%3] )
      {
        l2 = 1. / (_h[hMin%3] * _h[hMin%3]);
        l3 = 1. / (_h[(hMin+1)%3] * _h[(hMin+1)%3]);
        hMin  = hMin%3;
      }
    else
      {
        l2 = 1. / (_h[(hMin+1)%3] * _h[(hMin+1)%3]);
        l3 = 1. / (_h[hMin%3] * _h[hMin%3]);
        hMin = (hMin+1)%3;
      }

    // project dir 1 so that it is perpendicular to dir 0
    double cosa = t1(0) * dirs[hMin][0] + t1(1) * dirs[hMin][1] + t1(2) * dirs[hMin][2];
    double vec[3];
    for( int i=0; i<3; i++ ) vec[i] = dirs[hMin][i] - cosa * t1(i);
    normInv = 1. / sqrt ( vec[0] * vec[0] + vec[1] * vec[1] + vec[2] * vec[2] );
    for (int i=0; i<3; i++) t2(i) = vec[i] * normInv;
    
    // find last dir as the direction orthogonal to the two previous ones
    t3 = crossprod(t1,t2);

    // build the metric
    M = metric(l1,l2,l3,t1,t2,t3);
  }

  // -------------------------------------------------------------------
  AnisoMeshSize::AnisoMeshSize(double h):
    MeshSizeBase()
  {
    M = metric( 1./(h*h) );
  }

  // -------------------------------------------------------------------
  AnisoMeshSize::AnisoMeshSize(const AnisoMeshSize &pm):
    MeshSizeBase()
  {
    M = pm.M;
  }

  // -------------------------------------------------------------------
  MeshSizeType AnisoMeshSize::getType() const 
  {
    return ANISOTROPIC;
  }

  // -------------------------------------------------------------------
  void AnisoMeshSize::intersect(const pMSize pMS1, const pMSize pMS2)
  {
#ifdef _HAVE_GMSH_
    M = intersection(pMS1->getMetric(),pMS2->getMetric());
#else
#endif
  }

  // -------------------------------------------------------------------
  void AnisoMeshSize::interpolate(const pMSize pMS0, const pMSize pMS1, 
                                  double param)
  {
#ifdef _HAVE_GMSH_
    M = interpolation(pMS0->getMetric(),pMS1->getMetric(),param);
#else
#endif
  }

  // -------------------------------------------------------------------
  void AnisoMeshSize::interpolate(const pMSize pMS0, const pMSize pMS1, 
                                  const pMSize pMS2, double u, double v)
  {
#ifdef _HAVE_GMSH_
    M = interpolation(pMS0->getMetric(), pMS1->getMetric(), 
                      pMS2->getMetric(), u, v);
#else
#endif
  }

  // -------------------------------------------------------------------
  void AnisoMeshSize::interpolate(const pMSize pMS0, const pMSize pMS1, 
                                  const pMSize pMS2, const pMSize pMS3, 
                                  double u, double v, double w)
  {
#ifdef _HAVE_GMSH_
    M = interpolation(pMS0->getMetric(), pMS1->getMetric(), 
                      pMS2->getMetric(), pMS3->getMetric(), 
                      u, v, w);
#else
#endif
  }

  // -------------------------------------------------------------------
  double AnisoMeshSize::size(int i) const
  {
    gmshMat V = gmshMat(3,3);
    gmshVec S = gmshVec(3);
    M.eig(V,S);
    return ( 1. / sqrt( S(i) ) );
  }

  // -------------------------------------------------------------------
  void AnisoMeshSize::sizes(double _h[3]) const
  {
    gmshMat V = gmshMat(3,3);
    gmshVec S = gmshVec(3);
    M.eig(V,S);
    for (int i=0; i<3; i++) _h[i] = ( 1. / sqrt( S(i) ) );
  }

  // -------------------------------------------------------------------
  void AnisoMeshSize::direction(int i, double dir[3]) const 
  {
    gmshMat V = gmshMat(3,3); 
    gmshVec S = gmshVec(3);
    M.eig(V,S); // each column of V is a direction, S gives corresponding 1/h^2, not sort
    for (int iC=0; iC<3; iC++) dir[iC] = V(iC,i);
  }

  // -------------------------------------------------------------------
  void AnisoMeshSize::scale(int dir, double factor)
  {
    MAdMsgSgl::instance().error(__LINE__,__FILE__,"Not implemented");
  }

  // -------------------------------------------------------------------
  void AnisoMeshSize::scale(double factor)
  {
    MAdMsgSgl::instance().error(__LINE__,__FILE__,"Not implemented");
  }

  // -------------------------------------------------------------------
  double AnisoMeshSize::getMeanLength() const
  {
    gmshMat V = gmshMat(3,3);
    gmshVec S = gmshVec(3);
    M.eig(V,S);
    double mean = 0.;
    for (int i=0; i<3; i++) mean += 1. / sqrt( S(i) );
    return ( MAdTHIRD * mean );
  }

  // -------------------------------------------------------------------
  // get the square of the norm in the metric
  double AnisoMeshSize::normSq(const double vec[3]) const
  {
    vec3 theVec = vec3(vec[0],vec[1],vec[2]);
    return dot(theVec,M,theVec);
  }

  // -------------------------------------------------------------------
  // get the square of desired edge length along 'vec'
  double AnisoMeshSize::lengthSqInDir(const double vec[3]) const
  {
    vec3 theVec = vec3(vec[0],vec[1],vec[2]);
    theVec.normalize();
    return ( 1. / dot(theVec,M,theVec) );
  }

  // -------------------------------------------------------------------
  void AnisoMeshSize::print(string name) const
  {
    std::cout<<"Printing AnisoMeshSize \'"<<name<<"\' ("<<this<<")\n";
    M.print("Mesh size metric");
  }

  // -------------------------------------------------------------------

}
